"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import abc
import collections.abc
from . import common_pb2
import grpc
import grpc.aio
from . import pybridge_pb2
import typing

_T = typing.TypeVar("_T")

class _MaybeAsyncIterator(collections.abc.AsyncIterator[_T], collections.abc.Iterator[_T], metaclass=abc.ABCMeta): ...

class _ServicerContext(grpc.ServicerContext, grpc.aio.ServicerContext):  # type: ignore[misc, type-arg]
    ...

class PyBridgeStub:
    def __init__(self, channel: typing.Union[grpc.Channel, grpc.aio.Channel]) -> None: ...
    Healthcheck: grpc.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    Close: grpc.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    IirFilterDesign: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeIirFilterDesignRequest,
        pybridge_pb2.PyBridgeIirFilterDesignReply,
    ]

    FileOpen: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileOpenRequest,
        common_pb2.StandardReply,
    ]

    FileGetStatus: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileStandardRequest,
        pybridge_pb2.PyBridgeGetStatusReply,
    ]

    FileGetSpec: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileStandardRequest,
        pybridge_pb2.PyBridgeGetSpecReply,
    ]

    FileGetSignals: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeGetSignalsRequest,
        pybridge_pb2.PyBridgeGetSignalsReply,
    ]

    SpikesPushToExportFile: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.SpikesPushToExportReq,
        common_pb2.StandardReply,
    ]

    LoadMearecOnPybridge: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileOpenRequest,
        pybridge_pb2.MearecFileMeta,
    ]

    SpikesSendMearecHDF5: grpc.UnaryStreamMultiCallable[
        pybridge_pb2.MearecWaveformStreamRequest,
        pybridge_pb2.MearecWaveformChunk,
    ]

    SpikesCloseMearecHDF5: grpc.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    GetMearecTemplate: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecTemplateRequest,
        pybridge_pb2.MearecTemplate,
    ]

    GetMearecWaveforms: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecWaveformRequest,
        pybridge_pb2.MearecWaveformChunk,
    ]

    GetMearecSignals: grpc.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecSignalRequest,
        pybridge_pb2.MearecSignalReply,
    ]

    StreamingRequest: grpc.StreamUnaryMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]
    """template functions"""

    StreamingReply: grpc.UnaryStreamMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]

    StreamingBidirectional: grpc.StreamStreamMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]

class PyBridgeAsyncStub:
    Healthcheck: grpc.aio.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    Close: grpc.aio.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    IirFilterDesign: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeIirFilterDesignRequest,
        pybridge_pb2.PyBridgeIirFilterDesignReply,
    ]

    FileOpen: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileOpenRequest,
        common_pb2.StandardReply,
    ]

    FileGetStatus: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileStandardRequest,
        pybridge_pb2.PyBridgeGetStatusReply,
    ]

    FileGetSpec: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileStandardRequest,
        pybridge_pb2.PyBridgeGetSpecReply,
    ]

    FileGetSignals: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeGetSignalsRequest,
        pybridge_pb2.PyBridgeGetSignalsReply,
    ]

    SpikesPushToExportFile: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.SpikesPushToExportReq,
        common_pb2.StandardReply,
    ]

    LoadMearecOnPybridge: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.PyBridgeFileOpenRequest,
        pybridge_pb2.MearecFileMeta,
    ]

    SpikesSendMearecHDF5: grpc.aio.UnaryStreamMultiCallable[
        pybridge_pb2.MearecWaveformStreamRequest,
        pybridge_pb2.MearecWaveformChunk,
    ]

    SpikesCloseMearecHDF5: grpc.aio.UnaryUnaryMultiCallable[
        common_pb2.StandardRequest,
        common_pb2.StandardReply,
    ]

    GetMearecTemplate: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecTemplateRequest,
        pybridge_pb2.MearecTemplate,
    ]

    GetMearecWaveforms: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecWaveformRequest,
        pybridge_pb2.MearecWaveformChunk,
    ]

    GetMearecSignals: grpc.aio.UnaryUnaryMultiCallable[
        pybridge_pb2.MearecSignalRequest,
        pybridge_pb2.MearecSignalReply,
    ]

    StreamingRequest: grpc.aio.StreamUnaryMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]
    """template functions"""

    StreamingReply: grpc.aio.UnaryStreamMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]

    StreamingBidirectional: grpc.aio.StreamStreamMultiCallable[
        pybridge_pb2.Request,
        pybridge_pb2.Reply,
    ]

class PyBridgeServicer(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def Healthcheck(
        self,
        request: common_pb2.StandardRequest,
        context: _ServicerContext,
    ) -> typing.Union[common_pb2.StandardReply, collections.abc.Awaitable[common_pb2.StandardReply]]: ...

    @abc.abstractmethod
    def Close(
        self,
        request: common_pb2.StandardRequest,
        context: _ServicerContext,
    ) -> typing.Union[common_pb2.StandardReply, collections.abc.Awaitable[common_pb2.StandardReply]]: ...

    @abc.abstractmethod
    def IirFilterDesign(
        self,
        request: pybridge_pb2.PyBridgeIirFilterDesignRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.PyBridgeIirFilterDesignReply, collections.abc.Awaitable[pybridge_pb2.PyBridgeIirFilterDesignReply]]: ...

    @abc.abstractmethod
    def FileOpen(
        self,
        request: pybridge_pb2.PyBridgeFileOpenRequest,
        context: _ServicerContext,
    ) -> typing.Union[common_pb2.StandardReply, collections.abc.Awaitable[common_pb2.StandardReply]]: ...

    @abc.abstractmethod
    def FileGetStatus(
        self,
        request: pybridge_pb2.PyBridgeFileStandardRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.PyBridgeGetStatusReply, collections.abc.Awaitable[pybridge_pb2.PyBridgeGetStatusReply]]: ...

    @abc.abstractmethod
    def FileGetSpec(
        self,
        request: pybridge_pb2.PyBridgeFileStandardRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.PyBridgeGetSpecReply, collections.abc.Awaitable[pybridge_pb2.PyBridgeGetSpecReply]]: ...

    @abc.abstractmethod
    def FileGetSignals(
        self,
        request: pybridge_pb2.PyBridgeGetSignalsRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.PyBridgeGetSignalsReply, collections.abc.Awaitable[pybridge_pb2.PyBridgeGetSignalsReply]]: ...

    @abc.abstractmethod
    def SpikesPushToExportFile(
        self,
        request: pybridge_pb2.SpikesPushToExportReq,
        context: _ServicerContext,
    ) -> typing.Union[common_pb2.StandardReply, collections.abc.Awaitable[common_pb2.StandardReply]]: ...

    @abc.abstractmethod
    def LoadMearecOnPybridge(
        self,
        request: pybridge_pb2.PyBridgeFileOpenRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.MearecFileMeta, collections.abc.Awaitable[pybridge_pb2.MearecFileMeta]]: ...

    @abc.abstractmethod
    def SpikesSendMearecHDF5(
        self,
        request: pybridge_pb2.MearecWaveformStreamRequest,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[pybridge_pb2.MearecWaveformChunk], collections.abc.AsyncIterator[pybridge_pb2.MearecWaveformChunk]]: ...

    @abc.abstractmethod
    def SpikesCloseMearecHDF5(
        self,
        request: common_pb2.StandardRequest,
        context: _ServicerContext,
    ) -> typing.Union[common_pb2.StandardReply, collections.abc.Awaitable[common_pb2.StandardReply]]: ...

    @abc.abstractmethod
    def GetMearecTemplate(
        self,
        request: pybridge_pb2.MearecTemplateRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.MearecTemplate, collections.abc.Awaitable[pybridge_pb2.MearecTemplate]]: ...

    @abc.abstractmethod
    def GetMearecWaveforms(
        self,
        request: pybridge_pb2.MearecWaveformRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.MearecWaveformChunk, collections.abc.Awaitable[pybridge_pb2.MearecWaveformChunk]]: ...

    @abc.abstractmethod
    def GetMearecSignals(
        self,
        request: pybridge_pb2.MearecSignalRequest,
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.MearecSignalReply, collections.abc.Awaitable[pybridge_pb2.MearecSignalReply]]: ...

    @abc.abstractmethod
    def StreamingRequest(
        self,
        request_iterator: _MaybeAsyncIterator[pybridge_pb2.Request],
        context: _ServicerContext,
    ) -> typing.Union[pybridge_pb2.Reply, collections.abc.Awaitable[pybridge_pb2.Reply]]:
        """template functions"""

    @abc.abstractmethod
    def StreamingReply(
        self,
        request: pybridge_pb2.Request,
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[pybridge_pb2.Reply], collections.abc.AsyncIterator[pybridge_pb2.Reply]]: ...

    @abc.abstractmethod
    def StreamingBidirectional(
        self,
        request_iterator: _MaybeAsyncIterator[pybridge_pb2.Request],
        context: _ServicerContext,
    ) -> typing.Union[collections.abc.Iterator[pybridge_pb2.Reply], collections.abc.AsyncIterator[pybridge_pb2.Reply]]: ...

def add_PyBridgeServicer_to_server(servicer: PyBridgeServicer, server: typing.Union[grpc.Server, grpc.aio.Server]) -> None: ...



"""Base client for gRPC services."""

from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from radiens_core._transport._auth import AuthInterceptor
from radiens_core._transport._channel_pool import ChannelPool

if TYPE_CHECKING:
    from types import TracebackType

    from radiens_core.models.common import ServiceEndpoint

T = TypeVar("T", bound="BaseClient")


class BaseClient:
    """Base class for all Radiens gRPC clients.

    Handles channel pooling, authentication interceptors, and common
    lifecycle management.
    """

    def __init__(
        self,
        addresses: dict[str, str],
        auth_server_address: str,
        fallback_service: ServiceEndpoint,
    ) -> None:
        """Initialize the base client.

        Parameters
        ----------
        addresses
            Map of service name to "host:port" address.
        auth_server_address
            Address used by AuthInterceptor for server UUID fallback.
        fallback_service
            The default service endpoint to use if a requested service is
            not found in the address map.
        """
        self._addresses = addresses
        self._fallback_service = fallback_service
        self._pool = ChannelPool(
            interceptors=[AuthInterceptor(server_address=auth_server_address)],
        )

    # --- Lifecycle ---

    def close(self) -> None:
        """Close all gRPC channels managed by this client."""
        self._pool.close()

    def __enter__(self: T) -> T:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.close()

    # --- Internal helpers ---

    def _server_address(self, service: ServiceEndpoint) -> str:
        """Resolve the server address for a given service.

        Falls back to the configured ``fallback_service`` if the
        requested service name is not in the address map.
        """
        return self._addresses.get(service, self._addresses[self._fallback_service])



"""Common enums not tied to a specific domain model."""

from enum import StrEnum


class ClientType(StrEnum):
    """Client type identifier."""

    ALLEGO = "AllegoClient"
    VIDERE = "VidereClient"
    CURATE = "CurateClient"


class ServiceEndpoint(StrEnum):
    """gRPC service endpoint identifier.

    Mirrors the proto ``GrpcServiceType`` enum values used as keys in
    the ``ServerSpec.service`` map returned by ``GetRadiensServers``.
    """

    ALLEGO_CORE = "ALLEGO_CORE"
    ALLEGO_PCACHE = "ALLEGO_PCACHE"
    ALLEGO_KPI = "ALLEGO_KPI"
    ALLEGO_NEURONS = "ALLEGO_NEURONS"
    RADIENS_CORE = "RADIENS_CORE"
    RADIENS_SPIKE_SORTER = "RADIENS_SPIKE_SORTER"
    RADIENS_DEV = "RADIENS_DEV"
    RADIENS_DASHBOARDS = "RADIENS_DASHBOARDS"
    ALLEGO_LIFECYCLE = "ALLEGO_LIFECYCLE"
    RADIENS_LIFECYCLE = "RADIENS_LIFECYCLE"



"""Spike data proto <-> domain converters."""

from __future__ import annotations

from collections import defaultdict
from typing import TYPE_CHECKING

import numpy as np
import numpy.typing as npt

from radiens_core.models.spikes import (
    ChannelSpikeData,
    NeuronInfo,
    NeuronSpikeData,
    NeuronsResult,
    SpikesSpec,
)

if TYPE_CHECKING:
    from radiens_core._generated import biointerface_pb2, spikesorter_pb2

# ---------------------------------------------------------------------------
# Internal helpers
# ---------------------------------------------------------------------------


def _timestamps_to_sec(
    ts_ints: list[int],
    time_range: list[float],
    ts_range: list[int],
) -> npt.NDArray[np.float64]:
    """Convert a list of hardware int64 timestamps to seconds.

    Uses linear interpolation anchored by the ``timeRange`` and
    ``timeStampRange`` fields from the raster reply.  Returns an array of
    zeros if the ranges are degenerate (empty or zero span).
    """
    n = len(ts_ints)
    if n == 0:
        return np.empty(0, dtype=np.float64)
    if len(time_range) < 2 or len(ts_range) < 2:
        return np.zeros(n, dtype=np.float64)
    span_ts = ts_range[1] - ts_range[0]
    if span_ts == 0:
        return np.full(n, time_range[0], dtype=np.float64)
    ts_arr = np.array(ts_ints, dtype=np.int64)
    scale = (time_range[1] - time_range[0]) / span_ts
    return np.asarray(time_range[0] + (ts_arr - ts_range[0]) * scale, dtype=np.float64)


def _collect_dataIdxs_by_label(
    raster: spikesorter_pb2.SpikeSorterRasterDataReply,
) -> dict[int, list[int]]:
    """Build a mapping from label → list of waveform row indices.

    Each ``SpikeTimestamp.dataIdxs`` contains the row index into the dense
    waveform matrix (``SpikesGetSpikesDense``) for that spike.  Takes the
    first element of ``dataIdxs`` per spike (the primary waveform row).
    """
    by_label: dict[int, list[int]] = defaultdict(list)
    for st in raster.spikeTimestamps:
        if len(st.dataIdxs) > 0:
            by_label[st.label].append(st.dataIdxs[0])
    return dict(by_label)


# ---------------------------------------------------------------------------
# SpikesSpec
# ---------------------------------------------------------------------------


def spikes_spec_from_proto(
    proto: biointerface_pb2.SpikesSpecReply,
) -> SpikesSpec:
    """Convert ``SpikesSpecReply`` proto to domain ``SpikesSpec``."""
    if proto.HasField("tR") and len(proto.tR.timeRangeSec) >= 2:
        time_range: tuple[float, float] = (
            proto.tR.timeRangeSec[0],
            proto.tR.timeRangeSec[1],
        )
    else:
        time_range = (0.0, 0.0)

    if proto.HasField("spikeTR") and len(proto.spikeTR.timeRangeSec) >= 2:
        spike_time_range: tuple[float, float] = (
            proto.spikeTR.timeRangeSec[0],
            proto.spikeTR.timeRangeSec[1],
        )
    else:
        spike_time_range = (0.0, 0.0)

    return SpikesSpec(
        biointerface_id=proto.biointerfaceID,
        dataset_uid=proto.datasetUID,
        num_neurons=proto.numNeurons,
        num_sites=proto.numSites,
        enabled_channel_indices=list(proto.enabledSiteNtvChanIdx),
        time_range=time_range,
        spike_time_range=spike_time_range,
    )


# ---------------------------------------------------------------------------
# Channel spike data (RasterMode.CHANNELS)
# ---------------------------------------------------------------------------


def channel_spike_data_from_raster(
    proto: spikesorter_pb2.SpikeSorterRasterDataReply,
) -> dict[int, ChannelSpikeData]:
    """Convert a CHANNELS-mode raster reply to per-channel spike data.

    Groups ``SpikeTimestamp`` objects by ``label`` (= native channel index in
    CHANNELS mode).  Converts hardware int64 timestamps to seconds using the
    linear interpolation range embedded in the reply.  Waveforms are always
    ``None``; call :func:`merge_waveforms_channel` to attach them.

    Args:
        proto: ``SpikeSorterRasterDataReply`` from a CHANNELS-mode request.

    Returns:
        Dict mapping native channel index to :class:`ChannelSpikeData`.
    """
    time_range = list(proto.timeRange)
    ts_range = list(proto.timeStampRange)

    # Group spike timestamps and labels by channel label
    chan_ts: dict[int, list[int]] = defaultdict(list)
    chan_labels: dict[int, list[int]] = defaultdict(list)
    for st in proto.spikeTimestamps:
        chan_ts[st.label].append(st.spikeTimestamp)
        chan_labels[st.label].append(st.label)

    result: dict[int, ChannelSpikeData] = {}
    for chan_idx, ts_ints in chan_ts.items():
        timestamps = _timestamps_to_sec(ts_ints, time_range, ts_range)
        labels = np.array(chan_labels[chan_idx], dtype=np.int32)
        result[chan_idx] = ChannelSpikeData(
            ntv_chan_idx=chan_idx,
            timestamps=timestamps,
            labels=labels,
            waveforms=None,
        )
    return result


def merge_waveforms_channel(
    result: dict[int, ChannelSpikeData],
    raster: spikesorter_pb2.SpikeSorterRasterDataReply,
    dense: biointerface_pb2.SpikesSpikeDataDenseReply,
) -> dict[int, ChannelSpikeData]:
    """Attach waveforms to channel spike data using the dense waveform matrix.

    Uses ``SpikeTimestamp.dataIdxs`` from the raster reply to index into the
    float32 dense matrix (shape ``[totalNSpikes, waveformNPoints]``).

    Args:
        result: Channel spike data returned by :func:`channel_spike_data_from_raster`.
        raster: The same ``SpikeSorterRasterDataReply`` used to build *result*.
        dense: ``SpikesSpikeDataDenseReply`` from ``SpikesGetSpikesDense``.

    Returns:
        New dict with waveforms filled in for each channel present in both
        *result* and the dense reply.  Channels with no waveform rows are
        returned unchanged (waveforms remain ``None``).
    """
    total = dense.totalNSpikes
    n_pts = dense.waveformNPoints
    if total == 0 or n_pts == 0 or len(dense.data) == 0:
        return result

    wfm = np.frombuffer(dense.data, dtype=np.float32).reshape(total, n_pts)
    by_label = _collect_dataIdxs_by_label(raster)

    updated: dict[int, ChannelSpikeData] = {}
    for chan_idx, csd in result.items():
        row_indices = by_label.get(chan_idx)
        if row_indices and len(row_indices) > 0:
            waveforms: npt.NDArray[np.float32] | None = wfm[np.array(row_indices, dtype=np.intp)]
        else:
            waveforms = None
        updated[chan_idx] = ChannelSpikeData(
            ntv_chan_idx=csd.ntv_chan_idx,
            timestamps=csd.timestamps,
            labels=csd.labels,
            waveforms=waveforms,
        )
    return updated


# ---------------------------------------------------------------------------
# Neuron spike data (RasterMode.NEURONS)
# ---------------------------------------------------------------------------


def neuron_spike_data_from_raster(
    proto: spikesorter_pb2.SpikeSorterRasterDataReply,
    label_to_chan: dict[int, int],
) -> dict[int, NeuronSpikeData]:
    """Convert a NEURONS-mode raster reply to per-neuron spike data.

    Groups ``SpikeTimestamp`` objects by ``label`` (= integer neuron label in
    NEURONS mode).  Converts hardware int64 timestamps to seconds.  The
    parent channel for each neuron is looked up in *label_to_chan*; if the
    label is not found, ``ntv_chan_idx`` is set to ``-1``.

    Args:
        proto: ``SpikeSorterRasterDataReply`` from a NEURONS-mode request.
        label_to_chan: Mapping of integer neuron label → native channel index,
            built from ``SpikesSpecReply.site`` by the caller.

    Returns:
        Dict mapping integer neuron label to :class:`NeuronSpikeData`.
    """
    time_range = list(proto.timeRange)
    ts_range = list(proto.timeStampRange)

    neuron_ts: dict[int, list[int]] = defaultdict(list)
    for st in proto.spikeTimestamps:
        neuron_ts[st.label].append(st.spikeTimestamp)

    result: dict[int, NeuronSpikeData] = {}
    for label, ts_ints in neuron_ts.items():
        timestamps = _timestamps_to_sec(ts_ints, time_range, ts_range)
        result[label] = NeuronSpikeData(
            label=label,
            ntv_chan_idx=label_to_chan.get(label, -1),
            timestamps=timestamps,
            waveforms=None,
        )
    return result


def merge_waveforms_neuron(
    result: dict[int, NeuronSpikeData],
    raster: spikesorter_pb2.SpikeSorterRasterDataReply,
    dense: biointerface_pb2.SpikesSpikeDataDenseReply,
) -> dict[int, NeuronSpikeData]:
    """Attach waveforms to neuron spike data using the dense waveform matrix.

    Uses ``SpikeTimestamp.dataIdxs`` from the raster reply to index into the
    float32 dense matrix (shape ``[totalNSpikes, waveformNPoints]``).

    Args:
        result: Neuron spike data returned by :func:`neuron_spike_data_from_raster`.
        raster: The same ``SpikeSorterRasterDataReply`` used to build *result*.
        dense: ``SpikesSpikeDataDenseReply`` from ``SpikesGetSpikesDense``.

    Returns:
        New dict with waveforms filled in for each neuron present in both
        *result* and the dense reply.  Neurons with no waveform rows are
        returned unchanged (waveforms remain ``None``).
    """
    total = dense.totalNSpikes
    n_pts = dense.waveformNPoints
    if total == 0 or n_pts == 0 or len(dense.data) == 0:
        return result

    wfm = np.frombuffer(dense.data, dtype=np.float32).reshape(total, n_pts)
    by_label = _collect_dataIdxs_by_label(raster)

    updated: dict[int, NeuronSpikeData] = {}
    for label, nsd in result.items():
        row_indices = by_label.get(label)
        if row_indices and len(row_indices) > 0:
            waveforms: npt.NDArray[np.float32] | None = wfm[np.array(row_indices, dtype=np.intp)]
        else:
            waveforms = None
        updated[label] = NeuronSpikeData(
            label=nsd.label,
            ntv_chan_idx=nsd.ntv_chan_idx,
            timestamps=nsd.timestamps,
            waveforms=waveforms,
        )
    return updated


# ---------------------------------------------------------------------------
# NeuronInfo / NeuronsResult
# ---------------------------------------------------------------------------


def neuron_info_from_proto(
    pb: biointerface_pb2.NeuronDescriptor,
) -> NeuronInfo:
    """Convert a ``NeuronDescriptor`` proto to domain ``NeuronInfo``."""
    return NeuronInfo(
        neuron_id=pb.neuronID,
        site_ntv_chan_idx=pb.siteNtvChanIdx,
        spike_label=pb.spikeLabel,
        spike_count=pb.spikeCount,
        spike_rate=pb.spikeRate,
        mean_abs_peak_wfm=pb.meanAbsPeakWfm,
        snr=pb.snr,
        pos_probe=list(pb.posProbe),
        pos_tissue=list(pb.posTissue),
    )


def neurons_result_from_proto(
    pb: biointerface_pb2.BiointerfaceGetNeuronsReply,
) -> NeuronsResult:
    """Convert ``BiointerfaceGetNeuronsReply`` proto to domain ``NeuronsResult``.

    The proto ``site`` field is a ``MessageMap[int, NeuronL]`` keyed by
    native channel index.  Each ``NeuronL.neuron`` list contains the neuron
    descriptors for that channel.
    """
    neurons_by_channel: dict[int, list[NeuronInfo]] = {}
    for ntv_chan_idx, neuron_l in pb.site.items():
        neurons_by_channel[ntv_chan_idx] = [neuron_info_from_proto(nd) for nd in neuron_l.neuron]
    return NeuronsResult(neurons_by_channel=neurons_by_channel)



"""Print signal shapes from the live cache every 500 ms."""

from __future__ import annotations

import time

from radiens_core import AllegoClient
from radiens_core.models.dataset import TimeRangeSpec

LOOKBACK_SEC = 1.0
INTERVAL_SEC = 0.5

with AllegoClient() as client:
    print(f"Polling every {INTERVAL_SEC * 1000:.0f} ms — lookback {LOOKBACK_SEC} s. Ctrl-C to stop.\n")
    while True:
        sigs = client.get_signals(TimeRangeSpec.lookback(LOOKBACK_SEC), signals="all")
        amp = sigs.amp.shape if sigs.amp is not None else None
        ain = sigs.gpio_ain.shape if sigs.gpio_ain is not None else None
        din = sigs.gpio_din.shape if sigs.gpio_din is not None else None
        dout = sigs.gpio_dout.shape if sigs.gpio_dout is not None else None
        print(f"amp={amp}  ain={ain}  din={din}  dout={dout}")
        time.sleep(INTERVAL_SEC)



"""DIO (Digital Output) converters — proto <-> domain."""

from __future__ import annotations

from radiens_core._generated import allegoserver_pb2
from radiens_core.models.dio import DigitalOutChannelRegister, DigitalOutRegister, DIOMode

# --- DIOMode ---

_DIO_MODE_TO_PROTO: dict[DIOMode, allegoserver_pb2.DIOMode.ValueType] = {
    DIOMode.MANUAL: allegoserver_pb2.manual,
    DIOMode.EVENTS: allegoserver_pb2.events,
    DIOMode.GATED: allegoserver_pb2.gated,
    DIOMode.PULSE: allegoserver_pb2.pulse,
}

_DIO_MODE_FROM_PROTO: dict[int, DIOMode] = {
    int(allegoserver_pb2.manual): DIOMode.MANUAL,
    int(allegoserver_pb2.events): DIOMode.EVENTS,
    int(allegoserver_pb2.gated): DIOMode.GATED,
    int(allegoserver_pb2.pulse): DIOMode.PULSE,
}


def dio_mode_to_proto(mode: DIOMode) -> allegoserver_pb2.DIOMode.ValueType:
    """Convert domain DIOMode to proto DIOMode."""
    return _DIO_MODE_TO_PROTO[mode]


def dio_mode_from_proto(pb: int) -> DIOMode:
    """Convert proto DIOMode int to domain DIOMode."""
    return _DIO_MODE_FROM_PROTO[pb]


# --- Requests ---


def set_dio_events_to_proto(
    chan_idx: int,
    threshold: float,
    polarity: bool,
) -> allegoserver_pb2.DIOModeEventsRequest:
    """Convert DIO events parameters to ``DIOModeEventsRequest`` proto."""
    return allegoserver_pb2.DIOModeEventsRequest(
        chanIdx=chan_idx,
        eventThreshold=threshold,
        eventPolarity=polarity,
    )


def set_dio_manual_to_proto(
    chan_idx: int,
    state: bool,
) -> allegoserver_pb2.DIOModeManualRequest:
    """Convert DIO manual parameters to ``DIOModeManualRequest`` proto."""
    return allegoserver_pb2.DIOModeManualRequest(
        chanIdx=chan_idx,
        state=state,
    )


def set_dio_gated_to_proto(
    polarity: bool,
) -> allegoserver_pb2.DIOModeGatedRequest:
    """Convert DIO gated parameters to ``DIOModeGatedRequest`` proto."""
    return allegoserver_pb2.DIOModeGatedRequest(
        gatedPolarity=polarity,
    )


def set_dio_pulse_to_proto(
    interval_sec: float,
    duration_sec: float,
    chan_idx: int = -1,
) -> allegoserver_pb2.DIOModePulseRequest:
    """Convert DIO pulse parameters to ``DIOModePulseRequest`` proto."""
    return allegoserver_pb2.DIOModePulseRequest(
        reqIntervalSec=interval_sec,
        reqDurationSec=duration_sec,
        reqNtvChanIdx=chan_idx,
    )


# --- Register ---


def digital_out_channel_register_from_proto(
    proto: allegoserver_pb2.DigitalOutChannelRegister,
) -> DigitalOutChannelRegister:
    """Convert ``DigitalOutChannelRegister`` proto to domain model."""
    return DigitalOutChannelRegister(
        manual_state=proto.manualState,
        event_threshold=proto.eventThreshold,
        event_polarity=proto.eventPolarity,
        ntv_chan_idx=proto.ntvChanIdx,
        pulse_interval_sec=proto.pulseIntervalSec,
        pulse_duration_sec=proto.pulseDurationSec,
    )


def digital_out_register_from_proto(
    proto: allegoserver_pb2.DigitalOutRegister,
) -> DigitalOutRegister:
    """Convert ``DigitalOutRegister`` proto to domain model."""
    return DigitalOutRegister(
        mode=dio_mode_from_proto(proto.mode),
        dout_chan_registers=[digital_out_channel_register_from_proto(c) for c in proto.doutChanRegisters],
        gated_polarity=proto.gatedPolarity,
        req_pulse_interval_sec=proto.reqPulseIntervalSec,
        req_pulse_duration_sec=proto.reqPulseDurationSec,
        req_pulse_ntv_chan_idx=proto.reqPulseNtvChanIdx,
    )



"""Core hardware configuration models."""

from __future__ import annotations

from pydantic import BaseModel, ConfigDict, Field

from radiens_core.models.ports import Port
from radiens_core.models.status import BackboneMode


class CableDelay(BaseModel):
    """Cable delay configuration for a specific port.

    Attributes:
        port: The hardware port for which the cable delay is configured. References Port.
        cable_delay: The cable delay value in units specific to the hardware.
        is_auto: Whether the cable delay is automatically determined by the system.
    """

    model_config = ConfigDict(frozen=True, populate_by_name=True)

    port: Port
    cable_delay: int = Field(alias="cableDelay")
    is_auto: bool = Field(alias="isAuto")


class CoreConfig(BaseModel):
    """Core hardware configuration retrieved from the server.

    Attributes:
        allego_core_server_port: The network port used by the Allego Core server.
        base_samp_freq: Base sampling frequency in Hertz.
        stream_loop_dur_ms: Stream loop duration in milliseconds.
        backbone_mode: Current hardware backbone mode. References BackboneMode.
        has_xdaq_gpio_expander: Whether the system is equipped with an XDAQ GPIO expander.
        cable_delays: Mapping of port indices to their respective CableDelay configurations.
    """

    model_config = ConfigDict(frozen=True, populate_by_name=True)

    allego_core_server_port: str = Field(alias="allegoCoreServerPort")
    base_samp_freq: float = Field(alias="baseSampFreq")
    stream_loop_dur_ms: int = Field(alias="streamLoopDurMs")
    backbone_mode: BackboneMode = Field(alias="backboneMode")
    has_xdaq_gpio_expander: bool = Field(alias="hasXDAQGPIOExpander")
    cable_delays: dict[int, CableDelay] = Field(alias="cableDelays")


class SetCoreConfigRequest(BaseModel):
    """Request to update core hardware configuration.

    Attributes:
        samp_freq: Target sampling frequency in Hertz. Defaults to None.
        loop_dur: Target loop duration in milliseconds. Defaults to None.
        pcache_persistence: PCache persistence duration. Defaults to None.
        cable_delay: Specific cable delay configuration to set. References CableDelay. Defaults to None.
    """

    model_config = ConfigDict(frozen=True, populate_by_name=True)

    samp_freq: float | None = Field(default=None, alias="sampFreq")
    loop_dur: float | None = Field(default=None, alias="loopDur")
    pcache_persistence: float | None = Field(default=None, alias="pcachePersistence")
    cable_delay: CableDelay | None = Field(default=None, alias="cableDelay")



"""Tests for record state converters."""

from radiens_core._converters._record_state import record_state_request_to_proto
from radiens_core.models.ports import Port
from radiens_core.models.status import RecordMode, RecordStateRequest


def test_record_state_request_to_proto_system_level() -> None:
    """Test conversion of system-level RecordStateRequest to proto."""
    # Test OFF mode (system level - no port specified)
    request_off = RecordStateRequest(mode=RecordMode.OFF)
    proto_off = record_state_request_to_proto(request_off)
    assert proto_off.mode == 0  # R_OFF proto value
    assert not proto_off.HasField("port")  # No port field set

    # Test ON mode (system level - no port specified)
    request_on = RecordStateRequest(mode=RecordMode.ON)
    proto_on = record_state_request_to_proto(request_on)
    assert proto_on.mode == 1  # R_ON proto value
    assert not proto_on.HasField("port")  # No port field set


def test_record_state_request_to_proto_per_port() -> None:
    """Test conversion of per-port RecordStateRequest to proto."""
    # Test with specific port
    request_port_a = RecordStateRequest(mode=RecordMode.ON, port=Port.A)
    proto_port_a = record_state_request_to_proto(request_port_a)
    assert proto_port_a.mode == 1  # R_ON proto value
    assert proto_port_a.HasField("port")  # Port field is set
    assert proto_port_a.port == 0  # Port A proto value

    # Test with different port
    request_port_h = RecordStateRequest(mode=RecordMode.OFF, port=Port.H)
    proto_port_h = record_state_request_to_proto(request_port_h)
    assert proto_port_h.mode == 0  # R_OFF proto value
    assert proto_port_h.HasField("port")  # Port field is set
    assert proto_port_h.port == 7  # Port H proto value



"""Server discovery — find running Radiens servers and resolve service ports.

Discovery flow:
1. OS-level scan (lsof / tasklist+netstat) finds radiensserver listener ports.
2. Probe each discovered port with ``RadiensLifecycleStub.Healthcheck`` to
   find the lifecycle port.
3. Call ``GetRadiensServers`` on the lifecycle port to get the full
   service→port mapping for every ``ServiceEndpoint``.

Nothing in this module is part of the public API.
"""

from __future__ import annotations

import logging
import re
import subprocess
import sys
from typing import TYPE_CHECKING

import grpc

from radiens_core._constants import (
    ALLEGO_CORE_PORT,
    KPI_PORT,
    NEURONS1_PORT,
    PCACHE_PORT,
)
from radiens_core._generated import common_pb2, radiensserver_pb2, radiensserver_pb2_grpc
from radiens_core.exceptions import ServerDiscoveryError

if TYPE_CHECKING:
    from radiens_core._transport._channel_pool import ChannelPool

_log = logging.getLogger(__name__)

# Ports reserved for allego — never returned by radiensserver discovery.
_ALLEGO_PORTS: frozenset[int] = frozenset({ALLEGO_CORE_PORT, PCACHE_PORT, KPI_PORT, NEURONS1_PORT})

_RADIENSSERVER_PROCESS_PATTERN: str = "radienss"


# ---------------------------------------------------------------------------
# 1. OS-level process scanning
# ---------------------------------------------------------------------------


def discover_radiensserver_processes() -> list[str]:
    """Find listening ``radiensserver`` addresses via OS utilities.

    Returns a list of ``"host:port"`` strings, excluding allego-reserved ports.
    Returns an empty list if no servers are found or the OS is unsupported.
    """
    if sys.platform in ("linux", "darwin"):
        return _discover_unix()
    if sys.platform == "win32":
        return _discover_windows()
    _log.warning("Server discovery not supported on platform: %s", sys.platform)
    return []


def _discover_unix() -> list[str]:
    """Use ``lsof -i -n -P`` to find radiensserver listener ports."""
    addresses: list[str] = []
    try:
        proc = subprocess.run(
            ["lsof", "-i", "-n", "-P"],
            capture_output=True,
            text=True,
            timeout=10,
        )
    except (FileNotFoundError, subprocess.TimeoutExpired):
        _log.debug("lsof unavailable or timed out")
        return []

    for line in proc.stdout.splitlines():
        if re.search(_RADIENSSERVER_PROCESS_PATTERN, line) is None:
            continue
        if re.search(r"\(LISTEN\)", line) is None:
            continue
        # Format: "... TCP <ip>:<port> (LISTEN)"
        tcp_match = re.search(r"TCP\s+(\S+):(\d+)\s+\(LISTEN\)", line)
        if tcp_match is None:
            continue
        ip_raw = tcp_match.group(1)
        port_str = tcp_match.group(2)
        try:
            port = int(port_str)
        except ValueError:
            continue
        if port in _ALLEGO_PORTS:
            continue
        ip = "localhost" if ip_raw == "*" else ip_raw
        addresses.append(f"{ip}:{port}")

    return addresses


def _discover_windows() -> list[str]:
    """Use ``tasklist`` + ``netstat`` to find radiensserver listener ports."""
    # Step 1: find PIDs of radiensserver.exe
    pids: list[int] = []
    try:
        proc = subprocess.run(
            ["tasklist.exe", "/FI", "IMAGENAME eq radiensserver.exe"],
            capture_output=True,
            text=True,
            timeout=10,
        )
    except (FileNotFoundError, subprocess.TimeoutExpired):
        _log.debug("tasklist unavailable or timed out")
        return []

    for line in proc.stdout.splitlines():
        if "radiensserver.exe" not in line.lower():
            continue
        pid_match = re.search(r"\d+", line)
        if pid_match is not None:
            pids.append(int(pid_match.group()))

    if not pids:
        return []

    # Step 2: find LISTENING ports for those PIDs
    addresses: list[str] = []
    try:
        proc = subprocess.run(
            ["netstat.exe", "-a", "-o", "-n", "-p", "TCP"],
            capture_output=True,
            text=True,
            timeout=10,
        )
    except (FileNotFoundError, subprocess.TimeoutExpired):
        _log.debug("netstat unavailable or timed out")
        return []

    for line in proc.stdout.splitlines():
        for pid in pids:
            if re.search(rf"LISTENING\s+{pid}\b", line) is None:
                continue
            port_match = re.search(r":(\d+)", line)
            if port_match is None:
                continue
            try:
                port = int(port_match.group(1))
            except ValueError:
                continue
            if port in _ALLEGO_PORTS:
                continue
            addresses.append(f"localhost:{port}")

    return addresses


# ---------------------------------------------------------------------------
# 2. Lifecycle port probing
# ---------------------------------------------------------------------------


def _find_lifecycle_address(pool: ChannelPool, candidates: list[str]) -> str | None:
    """Probe candidate addresses to find the RadiensLifecycle port.

    Tries a ``Healthcheck`` RPC on each candidate using the
    ``RadiensLifecycleStub``.  Returns the first address that responds,
    or ``None`` if none do.
    """

    for addr in candidates:
        try:
            channel = pool.get_channel(addr)
            stub = radiensserver_pb2_grpc.RadiensLifecycleStub(channel)
            stub.Healthcheck(
                radiensserver_pb2.RadiensHealthcheckRequest(),
                timeout=3,
            )
            return addr
        except grpc.RpcError:
            _log.debug("RadiensLifecycle probe failed on %s", addr)
            continue
    return None


# ---------------------------------------------------------------------------
# 3. RPC-based service port resolution
# ---------------------------------------------------------------------------


def _resolve_via_lifecycle(
    pool: ChannelPool,
    lifecycle_addr: str,
) -> dict[str, str]:
    """Call ``GetRadiensServers`` on a known lifecycle address.

    Returns a ``ServiceEndpoint``-keyed ``dict[str, str]`` mapping every
    discovered service to its ``"host:port"`` address.
    """

    host = lifecycle_addr.split(":")[0]
    channel = pool.get_channel(lifecycle_addr)
    stub = radiensserver_pb2_grpc.RadiensLifecycleStub(channel)
    reply = stub.GetRadiensServers(
        common_pb2.GetRadiensServersRequest(hostIPaddress=host),
    )

    result: dict[str, str] = {}

    for spec in reply.radiensserver:
        spec_host = spec.host or host
        for svc_name, grpc_svc in spec.service.items():
            result[svc_name] = f"{spec_host}:{grpc_svc.port}"

    return result


# ---------------------------------------------------------------------------
# 4. High-level discovery entry point
# ---------------------------------------------------------------------------


def discover_radiens_addresses(pool: ChannelPool) -> dict[str, str]:
    """Discover running radiens servers and resolve all service ports.

    1. OS-scan for radiensserver listener ports.
    2. Probe each to find the ``RadiensLifecycle`` port.
    3. Call ``GetRadiensServers`` to get the full service→port map.

    Returns
    -------
    dict[str, str]
        ``ServiceEndpoint`` value → ``"host:port"`` for every discovered service.

    Raises
    ------
    radiens_core.exceptions.ServerDiscoveryError
        If no running radiensserver processes are found or none respond.
    """

    candidates = discover_radiensserver_processes()
    if not candidates:
        raise ServerDiscoveryError("No running radiensserver processes found. Start the Radiens application.")

    lifecycle_addr = _find_lifecycle_address(pool, candidates)
    if lifecycle_addr is None:
        raise ServerDiscoveryError("Found radiensserver processes but none responded to RadiensLifecycle healthcheck.")

    try:
        result = _resolve_via_lifecycle(pool, lifecycle_addr)
    except grpc.RpcError:
        raise ServerDiscoveryError(
            f"RadiensLifecycle at {lifecycle_addr} responded to healthcheck but GetRadiensServers RPC failed."
        ) from None

    if not result:
        raise ServerDiscoveryError(f"GetRadiensServers on {lifecycle_addr} returned an empty service map.")

    return result



"""RadiensLifecycle RPC wrappers — thin service layer (proto only)."""

from __future__ import annotations

from typing import TYPE_CHECKING

import grpc

from radiens_core._generated import (
    common_pb2,
    radiensserver_pb2,
    radiensserver_pb2_grpc,
)
from radiens_core._transport._errors import handle_grpc_error
from radiens_core.models.common import ClientType, ServiceEndpoint

if TYPE_CHECKING:
    from radiens_core._transport._channel_pool import ChannelPool


def healthcheck_proto(
    pool: ChannelPool,
    addr: str,
    client_type: ClientType = ClientType.VIDERE,
    service: ServiceEndpoint = ServiceEndpoint.RADIENS_LIFECYCLE,
) -> radiensserver_pb2.RadiensHealthcheckSpec:
    """Send ``Healthcheck`` RPC to a radiens service, return raw proto response.

    *service* selects the stub: ``RADIENS_CORE`` uses ``RadiensCoreStub``,
    anything else defaults to ``RadiensLifecycleStub``.
    """
    channel = pool.get_channel(addr)
    req = radiensserver_pb2.RadiensHealthcheckRequest()
    try:
        if service == ServiceEndpoint.RADIENS_CORE:
            return radiensserver_pb2_grpc.RadiensCoreStub(channel).Healthcheck(req)
        return radiensserver_pb2_grpc.RadiensLifecycleStub(channel).Healthcheck(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, client_type)
        raise  # unreachable, but satisfies mypy


def get_radiens_servers_proto(
    pool: ChannelPool,
    addr: str,
    host_ip: str = "localhost",
    client_type: ClientType = ClientType.VIDERE,
) -> common_pb2.RadiensServersReply:
    """Send ``GetRadiensServers`` RPC to RadiensLifecycle, return raw proto response."""
    channel = pool.get_channel(addr)
    stub = radiensserver_pb2_grpc.RadiensLifecycleStub(channel)
    req = common_pb2.GetRadiensServersRequest(hostIPaddress=host_ip)
    try:
        return stub.GetRadiensServers(req)
    except grpc.RpcError as ex:
        handle_grpc_error(ex, client_type)
        raise  # unreachable, but satisfies mypy



"""Radiens Python client for neuroscience data acquisition and analysis."""

from importlib.metadata import version

__version__ = version("radiens-core")

from radiens_core.allego_client import AllegoClient
from radiens_core.curate_client import CurateClient
from radiens_core.models.common import ServiceEndpoint
from radiens_core.models.core_config import CableDelay, CoreConfig, SetCoreConfigRequest
from radiens_core.models.dataset import (
    BandParams,
    BulkSinkParams,
    BulkSourceParams,
    CARParams,
    DownsampleParams,
    FileInfo,
    HighLowPassParams,
    NotchParams,
    PairedRefParams,
    ProtocolSpec,
    SinkParams,
    SliceChannelsParams,
    SliceTimeParams,
    SourceParams,
    TransformEdge,
    TransformNode,
    TransformNodeType,
    VirtualRefParams,
)
from radiens_core.models.kpi import KpiMetric, KpiMetricId, KpiMetricsResult, KpiMode, KpiStatus
from radiens_core.models.spikes import ChannelSpikeData, NeuronInfo, NeuronSpikeData, NeuronsResult, SpikesSpec
from radiens_core.models.triggers import TriggerChannel
from radiens_core.videre_client import VidereClient

__all__ = [
    "AllegoClient",
    "BandParams",
    "BulkSinkParams",
    "BulkSourceParams",
    "CARParams",
    "CableDelay",
    "ChannelSpikeData",
    "CoreConfig",
    "CurateClient",
    "DownsampleParams",
    "FileInfo",
    "HighLowPassParams",
    "KpiMetric",
    "KpiMetricId",
    "KpiMetricsResult",
    "KpiMode",
    "KpiStatus",
    "NeuronInfo",
    "NeuronSpikeData",
    "NeuronsResult",
    "NotchParams",
    "PairedRefParams",
    "ProtocolSpec",
    "ServiceEndpoint",
    "SetCoreConfigRequest",
    "SinkParams",
    "SliceChannelsParams",
    "SliceTimeParams",
    "SourceParams",
    "SpikesSpec",
    "TransformEdge",
    "TransformNode",
    "TransformNodeType",
    "TriggerChannel",
    "VidereClient",
    "VirtualRefParams",
    "__version__",
]



"""Tests for spike data converters."""

from __future__ import annotations

import numpy as np
import pytest

from radiens_core._converters._spikes import (
    channel_spike_data_from_raster,
    merge_waveforms_channel,
    merge_waveforms_neuron,
    neuron_spike_data_from_raster,
    neurons_result_from_proto,
    spikes_spec_from_proto,
)
from radiens_core._generated import biointerface_pb2, common_pb2, spikesorter_pb2
from radiens_core.models.spikes import NeuronInfo, SpikesSpec

# ---------------------------------------------------------------------------
# Helpers
# ---------------------------------------------------------------------------


def _make_spike_timestamp(
    ts: int,
    label: int,
    data_idxs: list[int] | None = None,
) -> biointerface_pb2.SpikeTimestamp:
    return biointerface_pb2.SpikeTimestamp(
        spikeTimestamp=ts,
        label=label,
        dataIdxs=data_idxs if data_idxs is not None else [],
    )


def _make_raster_reply(
    spike_timestamps: list[biointerface_pb2.SpikeTimestamp],
    time_range: list[float],
    ts_range: list[int],
) -> spikesorter_pb2.SpikeSorterRasterDataReply:
    return spikesorter_pb2.SpikeSorterRasterDataReply(
        spikeTimestamps=spike_timestamps,
        timeRange=time_range,
        timeStampRange=ts_range,
    )


# ---------------------------------------------------------------------------
# SpikesSpec converter
# ---------------------------------------------------------------------------


class TestSpikesSpecConverter:
    def test_basic_fields(self) -> None:
        tr = common_pb2.TimeRange(timeRangeSec=[1.0, 5.0])
        spike_tr = common_pb2.TimeRange(timeRangeSec=[1.5, 4.5])
        proto = biointerface_pb2.SpikesSpecReply(
            biointerfaceID="bio-123",
            datasetUID="ds-uid",
            numNeurons=4,
            numSites=2,
            enabledSiteNtvChanIdx=[0, 1],
            tR=tr,
            spikeTR=spike_tr,
        )
        spec = spikes_spec_from_proto(proto)
        assert isinstance(spec, SpikesSpec)
        assert spec.biointerface_id == "bio-123"
        assert spec.dataset_uid == "ds-uid"
        assert spec.num_neurons == 4
        assert spec.num_sites == 2
        assert spec.enabled_channel_indices == [0, 1]
        assert spec.time_range == (1.0, 5.0)
        assert spec.spike_time_range == (1.5, 4.5)

    def test_missing_time_range_defaults_to_zero(self) -> None:
        proto = biointerface_pb2.SpikesSpecReply(
            biointerfaceID="bio-456",
            datasetUID="ds-uid2",
        )
        spec = spikes_spec_from_proto(proto)
        assert spec.time_range == (0.0, 0.0)
        assert spec.spike_time_range == (0.0, 0.0)


# ---------------------------------------------------------------------------
# channel_spike_data_from_raster
# ---------------------------------------------------------------------------


class TestChannelSpikeDataConverter:
    def test_groups_by_channel_label(self) -> None:
        """Spikes are grouped by their label (= channel index in CHANNELS mode)."""
        sts = [
            _make_spike_timestamp(100, label=0),
            _make_spike_timestamp(200, label=1),
            _make_spike_timestamp(300, label=0),
        ]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = channel_spike_data_from_raster(raster)

        assert set(result.keys()) == {0, 1}
        assert len(result[0].timestamps) == 2
        assert len(result[1].timestamps) == 1

    def test_timestamp_conversion(self) -> None:
        """Int64 timestamps are converted linearly to seconds."""
        # ts_range [0, 1000] maps to time_range [0.0, 1.0]
        # ts=500 → 0.5 s
        sts = [_make_spike_timestamp(500, label=0)]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = channel_spike_data_from_raster(raster)
        assert pytest.approx(result[0].timestamps[0]) == 0.5

    def test_labels_array_matches_timestamps(self) -> None:
        sts = [
            _make_spike_timestamp(100, label=0),
            _make_spike_timestamp(200, label=0),
        ]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = channel_spike_data_from_raster(raster)
        # label field for CHANNELS mode = channel index, so both spikes have label 0
        assert list(result[0].labels) == [0, 0]

    def test_waveforms_none_by_default(self) -> None:
        sts = [_make_spike_timestamp(100, label=0)]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = channel_spike_data_from_raster(raster)
        assert result[0].waveforms is None

    def test_empty_raster_returns_empty_dict(self) -> None:
        raster = _make_raster_reply([], time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = channel_spike_data_from_raster(raster)
        assert result == {}

    def test_degenerate_ts_range_returns_zeros(self) -> None:
        """Zero-span timestamp range produces zero timestamps."""
        sts = [_make_spike_timestamp(500, label=0)]
        raster = _make_raster_reply(sts, time_range=[1.0, 1.0], ts_range=[500, 500])
        result = channel_spike_data_from_raster(raster)
        assert result[0].timestamps[0] == pytest.approx(1.0)


# ---------------------------------------------------------------------------
# merge_waveforms_channel
# ---------------------------------------------------------------------------


class TestMergeWaveformsChannel:
    def _make_dense_reply(
        self,
        wfm_matrix: list[list[float]],
    ) -> biointerface_pb2.SpikesSpikeDataDenseReply:
        arr = np.array(wfm_matrix, dtype=np.float32)
        total, n_pts = arr.shape
        return biointerface_pb2.SpikesSpikeDataDenseReply(
            data=arr.tobytes(),
            totalNSpikes=total,
            waveformNPoints=n_pts,
        )

    def test_waveforms_attached_via_data_idxs(self) -> None:
        """dataIdxs links each spike to a row in the dense matrix."""
        # 3 waveform rows, 4 samples each
        wfm = [[1.0, 2.0, 3.0, 4.0], [5.0, 6.0, 7.0, 8.0], [9.0, 10.0, 11.0, 12.0]]
        dense = self._make_dense_reply(wfm)

        # chan 0 → spike at row 0, chan 1 → spike at row 2
        sts = [
            _make_spike_timestamp(100, label=0, data_idxs=[0]),
            _make_spike_timestamp(200, label=1, data_idxs=[2]),
        ]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        base_result = channel_spike_data_from_raster(raster)
        merged = merge_waveforms_channel(base_result, raster, dense)

        assert merged[0].waveforms is not None
        np.testing.assert_array_almost_equal(merged[0].waveforms[0], [1.0, 2.0, 3.0, 4.0])
        assert merged[1].waveforms is not None
        np.testing.assert_array_almost_equal(merged[1].waveforms[0], [9.0, 10.0, 11.0, 12.0])

    def test_empty_dense_leaves_waveforms_none(self) -> None:
        sts = [_make_spike_timestamp(100, label=0, data_idxs=[0])]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        base_result = channel_spike_data_from_raster(raster)
        dense = biointerface_pb2.SpikesSpikeDataDenseReply(
            data=b"",
            totalNSpikes=0,
            waveformNPoints=0,
        )
        merged = merge_waveforms_channel(base_result, raster, dense)
        assert merged[0].waveforms is None


# ---------------------------------------------------------------------------
# neuron_spike_data_from_raster
# ---------------------------------------------------------------------------


class TestNeuronSpikeDataConverter:
    def test_groups_by_neuron_label(self) -> None:
        """Spikes are grouped by neuron label (NEURONS mode)."""
        sts = [
            _make_spike_timestamp(100, label=1),
            _make_spike_timestamp(200, label=2),
            _make_spike_timestamp(300, label=1),
        ]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = neuron_spike_data_from_raster(raster, label_to_chan={1: 0, 2: 1})

        assert set(result.keys()) == {1, 2}
        assert len(result[1].timestamps) == 2
        assert len(result[2].timestamps) == 1

    def test_ntv_chan_idx_populated_from_map(self) -> None:
        sts = [_make_spike_timestamp(100, label=3)]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = neuron_spike_data_from_raster(raster, label_to_chan={3: 7})
        assert result[3].ntv_chan_idx == 7

    def test_ntv_chan_idx_defaults_to_minus_one_if_not_found(self) -> None:
        sts = [_make_spike_timestamp(100, label=99)]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = neuron_spike_data_from_raster(raster, label_to_chan={})
        assert result[99].ntv_chan_idx == -1

    def test_waveforms_none_by_default(self) -> None:
        sts = [_make_spike_timestamp(100, label=1)]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        result = neuron_spike_data_from_raster(raster, label_to_chan={1: 0})
        assert result[1].waveforms is None


class TestMergeWaveformsNeuron:
    def _make_dense_reply(
        self,
        wfm_matrix: list[list[float]],
    ) -> biointerface_pb2.SpikesSpikeDataDenseReply:
        arr = np.array(wfm_matrix, dtype=np.float32)
        total, n_pts = arr.shape
        return biointerface_pb2.SpikesSpikeDataDenseReply(
            data=arr.tobytes(),
            totalNSpikes=total,
            waveformNPoints=n_pts,
        )

    def test_waveforms_attached_by_neuron_label(self) -> None:
        wfm = [[1.0, 2.0], [3.0, 4.0]]
        dense = self._make_dense_reply(wfm)

        sts = [
            _make_spike_timestamp(100, label=1, data_idxs=[0]),
            _make_spike_timestamp(200, label=2, data_idxs=[1]),
        ]
        raster = _make_raster_reply(sts, time_range=[0.0, 1.0], ts_range=[0, 1000])
        base_result = neuron_spike_data_from_raster(raster, label_to_chan={1: 0, 2: 0})
        merged = merge_waveforms_neuron(base_result, raster, dense)

        assert merged[1].waveforms is not None
        np.testing.assert_array_almost_equal(merged[1].waveforms[0], [1.0, 2.0])
        assert merged[2].waveforms is not None
        np.testing.assert_array_almost_equal(merged[2].waveforms[0], [3.0, 4.0])


# ---------------------------------------------------------------------------
# NeuronsResult converter
# ---------------------------------------------------------------------------


class TestNeuronsResultConverter:
    def test_neurons_by_channel(self) -> None:
        nd = biointerface_pb2.NeuronDescriptor(
            neuronID="n1",
            siteNtvChanIdx=3,
            spikeLabel="1",
            spikeCount=100,
            spikeRate=10.0,
            meanAbsPeakWfm=50.0,
            snr=3.5,
            posProbe=[1.0, 2.0, 3.0],
            posTissue=[4.0, 5.0, 6.0],
        )
        neuron_l = biointerface_pb2.BiointerfaceGetNeuronsReply.NeuronL(neuron=[nd])
        reply = biointerface_pb2.BiointerfaceGetNeuronsReply(site={3: neuron_l})

        result = neurons_result_from_proto(reply)

        assert 3 in result.neurons_by_channel
        infos = result.neurons_by_channel[3]
        assert len(infos) == 1
        info = infos[0]
        assert isinstance(info, NeuronInfo)
        assert info.neuron_id == "n1"
        assert info.site_ntv_chan_idx == 3
        assert info.spike_label == "1"
        assert info.spike_count == 100
        assert info.spike_rate == pytest.approx(10.0)
        assert info.snr == pytest.approx(3.5)
        assert info.pos_probe == [1.0, 2.0, 3.0]
        assert info.pos_tissue == [4.0, 5.0, 6.0]

    def test_empty_reply_returns_empty_dict(self) -> None:
        reply = biointerface_pb2.BiointerfaceGetNeuronsReply()
        result = neurons_result_from_proto(reply)
        assert result.neurons_by_channel == {}
